from __future__ import annotations

import datetime
import uuid

from dateutil.relativedelta import relativedelta
from pydantic import BaseModel, Field


def _today() -> datetime.date:
    return datetime.date.today()


class ParameterChange(BaseModel):
    date: datetime.date
    field: str
    value: float
    uuid: str = Field(default_factory=lambda: uuid.uuid4().hex[:8])


class InputData(BaseModel):
    growth_rate: float
    spending_per_month: float
    inflation: float
    annual_salary_increase: float
    income_per_month: float
    extra_income: float
    # Above can also be in ParameterUpdate
    current_nw: float
    date_of_birth: datetime.date
    safe_withdraw_rate: float = 4
    post_fire_spending_per_month: float | None = (
        None  # Post-FIRE spending (e.g., for moving to cheaper country)
    )
    parameter_changes: list[ParameterChange] = []

    @property
    def now(self):
        return _today()

    def age_at(self, date: datetime.date) -> float:
        delta = relativedelta(date, self.date_of_birth)
        years = delta.years
        months = delta.months
        days = delta.days
        age_in_years = years + (months / 12) + (days / 365.25)
        return age_in_years

    @property
    def age(self) -> float:
        return self.age_at(self.now)

    @property
    def saving_per_month(self):
        return self.income_per_month + self.extra_income - self.spending_per_month

    @property
    def monthly_growth_rate(self):
        return (1 + self.growth_rate / 100) ** (1 / 12)

    @property
    def monthly_inflation(self):
        return (1 + self.inflation / 100) ** (1 / 12)

    @property
    def monthly_salary_increase_rate(self):
        return (1 + self.annual_salary_increase / 100) ** (1 / 12)


class Results(BaseModel):
    months: float
    nw: float
    income: float
    extra_income: float  # fixed extra income per month
    spending: float
    post_fire_spending: float | None = (
        None  # Post-FIRE target spending (inflation-adjusted)
    )
    delta_nw: float
    total_saved: float
    input_data: InputData

    @property
    def date(self):
        return self.input_data.now + datetime.timedelta(days=365.25 / 12) * self.months

    @property
    def age(self):
        return self.input_data.age_at(self.date)

    @property
    def years(self) -> float:
        return self.months / 12

    @property
    def safe_withdraw_rule_monthly(self) -> float:
        return self.safe_withdraw_rule_yearly / 12

    @property
    def saving(self) -> float:
        return self.income + self.extra_income - self.spending

    @property
    def safe_withdraw_rule_yearly(self) -> float:
        return self.nw * self.input_data.safe_withdraw_rate / 100

    @property
    def fire_spending_target(self) -> float:
        """The spending used for FIRE calculation (post-FIRE if set, else current)."""
        return (
            self.post_fire_spending
            if self.post_fire_spending is not None
            else self.spending
        )

    @property
    def safe_withdraw_minus_spending(self) -> float:
        """Difference between safe withdrawal and FIRE spending target."""
        return self.safe_withdraw_rule_monthly - self.fire_spending_target

    @property
    def is_fire_reached(self) -> bool:
        """Whether FIRE has been reached at this point."""
        return self.safe_withdraw_minus_spending >= 0

    @property
    def actual_spending(self) -> float:
        """Actual spending for display - switches to post-FIRE spending after FIRE."""
        if self.is_fire_reached and self.post_fire_spending is not None:
            return self.post_fire_spending
        return self.spending

    @property
    def investment_profits(self) -> float:
        return self.nw * self.input_data.monthly_growth_rate - self.nw

    @property
    def total_investment_profits(self) -> float:
        return self.nw - self.total_saved

    def next_month(self) -> "Results":
        while (
            self.input_data.parameter_changes
            and self.input_data.parameter_changes[0].date
            and self.date >= self.input_data.parameter_changes[0].date
        ):
            change = self.input_data.parameter_changes.pop(0)
            print(f"Changing {change.field} to {change.value} at {change.date}")
            if change.field in ("growth_rate", "inflation", "annual_salary_increase"):
                setattr(self.input_data, change.field, change.value)
            elif change.field == "income_per_month":
                self.income = change.value
            elif change.field == "extra_income":
                self.extra_income = change.value
            elif change.field == "spending_per_month":
                self.spending = change.value
            else:
                raise ValueError(f"Unknown field {change.field}")

        new_nw = self.nw + self.investment_profits + self.saving
        new_months = self.months + 1
        new_spending = self.spending * self.input_data.monthly_inflation
        new_post_fire_spending = (
            self.post_fire_spending * self.input_data.monthly_inflation
            if self.post_fire_spending is not None
            else None
        )
        new_income = self.income * self.input_data.monthly_salary_increase_rate
        new_delta_nw = new_nw - self.nw
        new_saved = self.total_saved + self.saving
        return Results(
            months=new_months,
            nw=new_nw,
            income=new_income,
            extra_income=self.extra_income,
            spending=new_spending,
            post_fire_spending=new_post_fire_spending,
            delta_nw=new_delta_nw,
            total_saved=new_saved,
            input_data=self.input_data,
        )


def retirement_index(results: list[Results]) -> int | None:
    for i, r in enumerate(results):
        if r.safe_withdraw_minus_spending >= 0:
            return i
    return None


def interpolate(start: float, end: float, fraction: float) -> float:
    return start + (end - start) * fraction


class Summary(BaseModel):
    age: float
    fire_age: float
    fire_date: datetime.date
    years_till_fi: float
    saving_at_fi: float
    safe_withdraw_at_fi: float
    spending_at_fi: float
    post_fire_spending_at_fi: float | None  # Post-FIRE spending target at FIRE (if set)
    nw_at_fi: float
    total_investment_profits: float
    total_saved: float
    safe_withdraw_at_age: dict[int, float]

    @classmethod
    def _interpolate_result(cls, results: list[Results], index: int) -> Results:
        if index == 0:
            return results[index]

        last = results[index]
        second_last = results[index - 1]

        fraction = (0 - second_last.safe_withdraw_minus_spending) / (
            last.safe_withdraw_minus_spending - second_last.safe_withdraw_minus_spending
        )

        interpolated_months = interpolate(second_last.months, last.months, fraction)

        post_fire_spending: float | None = None
        if (
            second_last.post_fire_spending is not None
            and last.post_fire_spending is not None
        ):
            post_fire_spending = interpolate(
                second_last.post_fire_spending, last.post_fire_spending, fraction
            )
        return Results(
            months=interpolated_months,
            nw=interpolate(second_last.nw, last.nw, fraction),
            income=interpolate(second_last.income, last.income, fraction),
            extra_income=second_last.extra_income,
            spending=interpolate(second_last.spending, last.spending, fraction),
            post_fire_spending=post_fire_spending,
            delta_nw=interpolate(second_last.delta_nw, last.delta_nw, fraction),
            total_saved=interpolate(
                second_last.total_saved, last.total_saved, fraction
            ),
            input_data=second_last.input_data,
        )

    @classmethod
    def from_results(cls, results: list[Results]) -> Summary | None:
        index = retirement_index(results)
        if index is None:
            return None

        r = cls._interpolate_result(results, index)

        safe_withdraw_at_age = {
            round(r.age): r.safe_withdraw_rule_yearly / 12
            for r in results
            if round(r.age, 1) % 1 == 0
        }
        return cls(
            age=r.input_data.age,
            fire_date=r.date,
            fire_age=r.age,
            years_till_fi=r.years,
            safe_withdraw_at_fi=r.safe_withdraw_rule_monthly,
            spending_at_fi=r.spending,
            post_fire_spending_at_fi=r.post_fire_spending,
            saving_at_fi=r.saving,
            nw_at_fi=r.nw,
            total_investment_profits=r.total_investment_profits,
            total_saved=r.total_saved,
            safe_withdraw_at_age=safe_withdraw_at_age,
        )


def calculate_results_for_month(
    data: InputData,
    target: int | datetime.date | None = None,
) -> list[Results]:
    # If target is a date, calculate the target month
    if isinstance(target, datetime.date):
        delta_months = (
            (target.year - data.now.year) * 12 + target.month - data.now.month
        )
    elif target is None:
        delta_months = 100 * 12
    else:
        delta_months = target

    # Set initial values
    r = Results(
        months=0,
        years=0,
        nw=data.current_nw,
        delta_nw=0,
        income=data.income_per_month,
        extra_income=data.extra_income,
        spending=data.spending_per_month,
        post_fire_spending=data.post_fire_spending_per_month,
        total_saved=data.current_nw,
        input_data=data,
    )
    results = [r]
    done_for = 0
    for _ in range(1, delta_months + 1):
        r = r.next_month()
        results.append(r)
        if r.safe_withdraw_minus_spending > 0:
            done_for += 1
            if done_for >= 6 * 12:  # 6 years
                break
    return results



from __future__ import annotations

import datetime
import uuid
from pathlib import Path
from typing import Optional
from urllib.parse import urlencode

from fastapi import FastAPI, Query, Request
from fastapi.responses import HTMLResponse, Response
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from fastapi_htmx import htmx, htmx_init

from .fire import InputData, ParameterChange, Summary, calculate_results_for_month
from .plots import (
    plot_age_vs_net_worth,
    plot_monthly_financial_flows,
)

FOLDER = Path(__file__).parent.resolve()


app = FastAPI()
app.mount("/static", StaticFiles(directory=FOLDER / "static"), name="static")
templates = Jinja2Templates(directory=FOLDER / "templates")
htmx_init(templates=templates)

# Default values for the input fields
DEFAULT_GROWTH_RATE = 7
DEFAULT_CURRENT_NW = 50_000
DEFAULT_SPENDING_PER_MONTH = 4_000
DEFAULT_INFLATION = 2
DEFAULT_ANNUAL_SALARY_INCREASE = 5
DEFAULT_INCOME_PER_MONTH = 8_000
DEFAULT_EXTRA_INCOME = 0
DEFAULT_DATE_OF_BIRTH = "1990-01-01"
DEFAULT_SAFE_WITHDRAW_RATE = 4
DEFAULT_EXTRA_SPENDING = 0
DEFAULT_POST_FIRE_SPENDING_PER_MONTH = 0  # 0 means use current spending

# Choices for parameter select boxes used across templates
PARAMETER_CHOICES: list[tuple[str, str]] = [
    ("growth_rate", "📈 Investment Growth Rate (%)"),
    ("spending_per_month", "🛒 Monthly Spending ($)"),
    ("inflation", "💰 Inflation Rate (%)"),
    ("annual_salary_increase", "📊 Annual Salary Increase (%)"),
    ("income_per_month", "💵 Monthly Income ($)"),
    ("extra_income", "💎 Extra Monthly Income ($)"),
]

# Expose frequently-used helpers & constants to every template once
# so we don't have to add them to each context dict.
templates.env.globals.update(
    parameter_choices=PARAMETER_CHOICES,
)


@app.get("/", response_class=HTMLResponse)
@htmx("index.html", "index.html")
async def index(
    request: Request,
    growth_rate: Optional[float] = DEFAULT_GROWTH_RATE,
    current_nw: Optional[float] = DEFAULT_CURRENT_NW,
    spending_per_month: Optional[float] = DEFAULT_SPENDING_PER_MONTH,
    inflation: Optional[float] = DEFAULT_INFLATION,
    annual_salary_increase: Optional[float] = DEFAULT_ANNUAL_SALARY_INCREASE,
    income_per_month: Optional[float] = DEFAULT_INCOME_PER_MONTH,
    extra_income: Optional[float] = DEFAULT_EXTRA_INCOME,
    date_of_birth: Optional[str] = DEFAULT_DATE_OF_BIRTH,
    safe_withdraw_rate: Optional[float] = DEFAULT_SAFE_WITHDRAW_RATE,
    extra_spending: Optional[float] = DEFAULT_EXTRA_SPENDING,
    post_fire_spending_per_month: Optional[
        float
    ] = DEFAULT_POST_FIRE_SPENDING_PER_MONTH,
    change_dates: list[str] = Query(default=[]),
    change_fields: list[str] = Query(default=[]),
    change_values: list[str] = Query(default=[]),
):
    parameter_changes = _parameter_changes(change_dates, change_fields, change_values)
    return {
        "request": request,
        "growth_rate": growth_rate,
        "current_nw": current_nw,
        "spending_per_month": spending_per_month,
        "inflation": inflation,
        "annual_salary_increase": annual_salary_increase,
        "income_per_month": income_per_month,
        "extra_income": extra_income,
        "date_of_birth": date_of_birth,
        "safe_withdraw_rate": safe_withdraw_rate,
        "extra_spending": extra_spending,
        "post_fire_spending_per_month": post_fire_spending_per_month,
        "parameter_changes": parameter_changes,
    }


@app.get("/add-parameter-change", response_class=HTMLResponse)
async def add_parameter_change(request: Request):
    unique_id = uuid.uuid4().hex[:8]  # Generate a unique identifier
    return templates.TemplateResponse(
        "parameter_change.html.jinja2", {"request": request, "uuid": unique_id}
    )


@app.get("/add-parameter-row/{uuid}", response_class=HTMLResponse)
async def add_parameter_row(request: Request, uuid: str, date: str = ""):
    """Add a new parameter row to an existing parameter change block"""
    return templates.TemplateResponse(
        "parameter_row.html.jinja2",
        {
            "request": request,
            "uuid": uuid,
            "field": "",
            "value": "",
            "date": date,
            "is_first_row": False,
        },
    )


@app.delete("/remove-parameter-change", response_class=HTMLResponse)
async def remove_parameter_change():
    # Since htmx handles the removal on the client side using hx-target and hx-swap,
    # the server doesn't need to perform any action. Just return an empty response.
    return Response("", status_code=200)


@app.delete("/remove-parameter-row", response_class=HTMLResponse)
async def remove_parameter_row():
    """Remove a parameter row from a parameter change block"""
    # Since htmx handles the removal on the client side using hx-target and hx-swap,
    # the server doesn't need to perform any action. Just return an empty response.
    return Response("", status_code=200)


def format_currency(value):
    """Format currency values in compact notation (e.g., $2.5M, $800k)"""
    if value < 1_000:
        return f"${value:.0f}"
    elif value < 100_000:
        return f"${value / 1_000:.1f}k"
    elif value < 1_000_000:
        return f"${value / 1_000:.0f}k"
    elif value < 10_000_000:
        return f"${value / 1_000_000:.1f}M"
    else:
        return f"${value / 1_000_000:.0f}M"


def interpolate_color(
    x: float,
    x_red: int = -2000,
    x_orange: int = -1000,
    x_yellow: int = 0,
    x_light_green: int = 1000,
    x_green: int = 2000,
    return_hex: bool = True,
):
    color_points = {
        x_red: (255, 0, 0),
        x_orange: (255, 167, 0),
        x_yellow: (255, 244, 0),
        x_light_green: (163, 255, 0),
        x_green: (44, 186, 0),
    }

    def _interpolate_color(c1, c2, ratio):
        return tuple(int(c1[i] + (c2[i] - c1[i]) * ratio) for i in range(3))

    if x <= x_red:
        rgb = color_points[x_red]
        return _rgb_to_hex(rgb) if return_hex else rgb

    elif x >= x_green:
        rgb = color_points[x_green]
        return _rgb_to_hex(rgb) if return_hex else rgb
    else:
        sorted_keys = sorted(color_points.keys())
        for i, key in enumerate(sorted_keys[:-1]):
            if key <= x < sorted_keys[i + 1]:
                c1 = color_points[key]
                c2 = color_points[sorted_keys[i + 1]]
                ratio = (x - key) / (sorted_keys[i + 1] - key)
                rgb = _interpolate_color(c1, c2, ratio)
                return _rgb_to_hex(rgb) if return_hex else rgb


def _rgb_to_hex(rgb: tuple[int, int, int]) -> str:
    return "#{:02x}{:02x}{:02x}".format(*rgb)


def _date_str_to_date(date_str: str) -> datetime.date:
    return datetime.datetime.strptime(date_str, "%Y-%m-%d").date()


def _parameter_changes(
    change_dates: list[str],
    change_fields: list[str],
    change_values: list[str],
) -> list[ParameterChange]:
    parameter_changes = []
    for date, field, value in zip(change_dates, change_fields, change_values):
        date_ = _date_str_to_date(date)
        parameter_changes.append(ParameterChange(date=date_, field=field, value=value))
    return sorted(parameter_changes, key=lambda x: x.date)


@app.get("/calculate", response_class=HTMLResponse)
@htmx("results_partial.html", "index.html")
async def calculate(
    request: Request,
    growth_rate: float = Query(default=DEFAULT_GROWTH_RATE),
    current_nw: float = Query(default=DEFAULT_CURRENT_NW),
    spending_per_month: float = Query(default=DEFAULT_SPENDING_PER_MONTH),
    inflation: float = Query(default=DEFAULT_INFLATION),
    annual_salary_increase: float = Query(default=DEFAULT_ANNUAL_SALARY_INCREASE),
    income_per_month: float = Query(default=DEFAULT_INCOME_PER_MONTH),
    extra_income: float = Query(default=DEFAULT_EXTRA_INCOME),
    date_of_birth: str = Query(default=DEFAULT_DATE_OF_BIRTH),
    safe_withdraw_rate: float = Query(default=DEFAULT_SAFE_WITHDRAW_RATE),
    extra_spending: float = Query(default=DEFAULT_EXTRA_SPENDING),
    post_fire_spending_per_month: float = Query(
        default=DEFAULT_POST_FIRE_SPENDING_PER_MONTH
    ),
    change_dates: list[str] = Query(default=[]),
    change_fields: list[str] = Query(default=[]),
    change_values: list[str] = Query(default=[]),
):
    parameter_changes = _parameter_changes(change_dates, change_fields, change_values)
    dob = _date_str_to_date(date_of_birth)
    # Convert post_fire_spending_per_month: 0 or empty means use current spending (None)
    post_fire_spending = (
        post_fire_spending_per_month if post_fire_spending_per_month > 0 else None
    )
    input_data = InputData(
        growth_rate=growth_rate,
        current_nw=current_nw,
        spending_per_month=spending_per_month,
        inflation=inflation,
        annual_salary_increase=annual_salary_increase,
        income_per_month=income_per_month,
        extra_income=extra_income,
        date_of_birth=dob,
        safe_withdraw_rate=safe_withdraw_rate,
        post_fire_spending_per_month=post_fire_spending,
        parameter_changes=parameter_changes,
    )
    input_data_with_extra = input_data.model_copy(
        update={"current_nw": current_nw - extra_spending}
    )

    # Calculate results without extra spending (main results)
    results = calculate_results_for_month(input_data)
    summary = Summary.from_results(results)

    # Calculate results with extra spending only for comparison
    results_with_extra = calculate_results_for_month(input_data_with_extra)
    summary_with_extra = Summary.from_results(results_with_extra)

    time_difference = None
    if summary and summary_with_extra:
        time_difference = (
            summary_with_extra.fire_date - summary.fire_date
        ).total_seconds() / (365.25 * 24 * 3600)

    if summary is not None:
        age_vs_net_worth_plot = plot_age_vs_net_worth(results, summary)
        monthly_financial_flows_plot = plot_monthly_financial_flows(results, summary)
    else:
        age_vs_net_worth_plot = None
        monthly_financial_flows_plot = None

    # Create URL parameters string
    url_params = urlencode(
        {
            "growth_rate": growth_rate,
            "current_nw": current_nw,
            "spending_per_month": spending_per_month,
            "inflation": inflation,
            "annual_salary_increase": annual_salary_increase,
            "income_per_month": income_per_month,
            "extra_income": extra_income,
            "date_of_birth": date_of_birth,
            "safe_withdraw_rate": safe_withdraw_rate,
            "extra_spending": extra_spending,
            "post_fire_spending_per_month": post_fire_spending_per_month,
            "change_dates": change_dates,
            "change_fields": change_fields,
            "change_values": change_values,
        }
    )

    context = {
        "request": request,
        "results": results,
        "summary": summary,
        "growth_rate": growth_rate,
        "current_nw": current_nw,
        "spending_per_month": spending_per_month,
        "inflation": inflation,
        "annual_salary_increase": annual_salary_increase,
        "income_per_month": income_per_month,
        "extra_income": extra_income,
        "date_of_birth": dob.strftime("%Y-%m-%d"),
        "safe_withdraw_rate": safe_withdraw_rate,
        "extra_spending": extra_spending,
        "post_fire_spending_per_month": post_fire_spending_per_month,
        "parameter_changes": parameter_changes,
        "age_vs_net_worth_plot": age_vs_net_worth_plot,
        "monthly_financial_flows_plot": monthly_financial_flows_plot,
        "time_difference": time_difference,
        "summary_with_extra": summary_with_extra,
        "url_params": url_params,
    }

    return context


# Register helper functions once they are defined so templates can access them
templates.env.globals.update(
    format_currency=format_currency,
    interpolate_color=interpolate_color,
)


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("app:app", host="127.0.0.1", port=8000, log_level="info", reload=True)



import datetime
from unittest.mock import patch

import pytest

from pydantic import ValidationError

import wenfire.fire
from wenfire.fire import (
    InputData,
    Results,
    Summary,
    calculate_results_for_month,
    interpolate,
)


# Fixtures
@pytest.fixture
def input_data() -> InputData:
    return InputData(
        growth_rate=5.0,
        current_nw=100000.0,
        spending_per_month=3000.0,
        inflation=2.0,
        annual_salary_increase=3.0,
        income_per_month=5000.0,
        extra_income=500.0,
        date_of_birth=datetime.date(1990, 1, 1),
        safe_withdraw_rate=4.0,
    )


# Mock today's date for consistent testing
@pytest.fixture(autouse=True)
def fixed_today():
    """
    A pytest fixture that mocks datetime.date.today() to return 2024-04-01.
    """
    fixed_date = datetime.date(2024, 4, 1)
    with patch.object(wenfire.fire, "_today", return_value=fixed_date):
        yield fixed_date


# Tests for InputData
def test_input_data_age_at(input_data: InputData) -> None:
    target_date = datetime.date(2024, 4, 1)
    age = input_data.age_at(target_date)
    expected_age = 34.25  # From 1990-01-01 to 2024-04-01
    assert abs(age - expected_age) < 0.01, f"Expected age {expected_age}, got {age}"


def test_input_data_age(input_data: InputData) -> None:
    age = input_data.age
    expected_age = 34.25  # From 1990-01-01 to 2024-04-01
    assert abs(age - expected_age) < 0.01, f"Expected age {expected_age}, got {age}"


def test_input_data_saving_per_month(input_data: InputData) -> None:
    expected_saving = (
        input_data.income_per_month
        + input_data.extra_income
        - input_data.spending_per_month
    )
    assert (
        input_data.saving_per_month == expected_saving
    ), "Incorrect saving_per_month calculation"


def test_input_data_monthly_growth_rate(input_data: InputData) -> None:
    expected = (1 + input_data.growth_rate / 100) ** (1 / 12)
    assert (
        abs(input_data.monthly_growth_rate - expected) < 1e-6
    ), "Incorrect monthly_growth_rate calculation"


def test_input_data_monthly_inflation(input_data: InputData) -> None:
    expected = (1 + input_data.inflation / 100) ** (1 / 12)
    assert (
        abs(input_data.monthly_inflation - expected) < 1e-6
    ), "Incorrect monthly_inflation calculation"


def test_input_data_monthly_salary_increase_rate(input_data: InputData) -> None:
    expected = (1 + input_data.annual_salary_increase / 100) ** (1 / 12)
    assert (
        abs(input_data.monthly_salary_increase_rate - expected) < 1e-6
    ), "Incorrect monthly_salary_increase_rate calculation"


# Tests for Results
def test_results_initial(input_data: InputData) -> None:
    r = Results(
        months=0,
        nw=input_data.current_nw,
        income=input_data.income_per_month,
        extra_income=input_data.extra_income,
        spending=input_data.spending_per_month,
        delta_nw=0.0,
        total_saved=input_data.current_nw,
        input_data=input_data,
    )
    assert r.months == 0
    assert r.nw == input_data.current_nw
    assert r.income == input_data.income_per_month
    assert r.extra_income == input_data.extra_income
    assert r.spending == input_data.spending_per_month
    assert r.delta_nw == 0.0
    assert r.total_saved == input_data.current_nw


def test_results_next_month(input_data: InputData) -> None:
    r = Results(
        months=0,
        nw=input_data.current_nw,
        income=input_data.income_per_month,
        extra_income=input_data.extra_income,
        spending=input_data.spending_per_month,
        delta_nw=0.0,
        total_saved=input_data.current_nw,
        input_data=input_data,
    )
    next_r = r.next_month()
    expected_nw = r.nw + r.investment_profits + r.saving
    expected_income = r.income * input_data.monthly_salary_increase_rate
    expected_spending = r.spending * input_data.monthly_inflation
    expected_delta_nw = expected_nw - r.nw
    expected_total_saved = r.total_saved + r.saving

    assert next_r.months == 1
    assert abs(next_r.nw - expected_nw) < 1e-6, "Incorrect nw after next_month"
    assert (
        abs(next_r.income - expected_income) < 1e-6
    ), "Incorrect income after next_month"
    assert (
        abs(next_r.spending - expected_spending) < 1e-6
    ), "Incorrect spending after next_month"
    assert (
        abs(next_r.delta_nw - expected_delta_nw) < 1e-6
    ), "Incorrect delta_nw after next_month"
    assert (
        abs(next_r.total_saved - expected_total_saved) < 1e-6
    ), "Incorrect total_saved after next_month"


# Tests for utility functions


def test_interpolate() -> None:
    assert interpolate(0, 10, 0.5) == 5.0, "Interpolation failed for middle fraction"
    assert (
        interpolate(10, 20, 0.25) == 12.5
    ), "Interpolation failed for quarter fraction"
    assert interpolate(-5, 5, 0.5) == 0.0, "Interpolation failed with negative numbers"
    assert interpolate(100, 200, 0) == 100, "Interpolation failed for fraction 0"
    assert interpolate(100, 200, 1) == 200, "Interpolation failed for fraction 1"


# Tests for Summary


def test_summary_from_results_no_retirement(input_data: InputData) -> None:
    # Create a list of Results where the retirement condition is never met
    r = Results(
        months=0,
        nw=100000.0,
        income=5000.0,
        extra_income=500.0,
        spending=3000.0,
        delta_nw=0.0,
        total_saved=100000.0,
        input_data=input_data,
    )
    results = [r] + [r.next_month() for _ in range(100)]
    summary = Summary.from_results(results)
    assert (
        summary is None
    ), "Summary should be None when retirement condition is not met"


# Tests for calculate_results_for_month
def test_calculate_results_for_month_no_target(input_data: InputData) -> None:
    results = calculate_results_for_month(input_data)
    assert len(results) > 0, "Results should not be empty"
    # Check first result
    first = results[0]
    assert first.months == 0
    assert first.nw == input_data.current_nw
    assert first.income == input_data.income_per_month
    assert first.spending == input_data.spending_per_month


def test_calculate_results_for_month_with_target_int(input_data: InputData) -> None:
    target_months = 24
    results = calculate_results_for_month(input_data, target=target_months)
    assert len(results) == 25, f"Expected 25 months, got {len(results)}"


def test_calculate_results_for_month_retirement_condition(
    input_data: InputData,
) -> None:
    # Modify input_data to meet retirement condition quickly
    input_data.spending_per_month = 4000.0  # Increase spending
    input_data.extra_income = 2000.0  # Increase extra income to make saving high

    results = calculate_results_for_month(input_data)
    # Expect retirement condition met early
    assert len(results) < 100 * 12, "Retirement condition should have been met early"


# Tests for Validation
def test_input_data_validation() -> None:
    with pytest.raises(ValidationError):
        InputData(
            growth_rate="five",  # Invalid type
            current_nw=100000.0,
            spending_per_month=3000.0,
            inflation=2.0,
            annual_salary_increase=3.0,
            income_per_month=5000.0,
            extra_income=500.0,
            date_of_birth="1990-01-01",  # Invalid type, should be datetime.date
        )


# Additional edge case tests
def test_age_at_leap_year(input_data: InputData) -> None:
    birth_date = datetime.date(2000, 2, 29)
    input_data.date_of_birth = birth_date
    target_date = datetime.date(2024, 2, 28)
    age = input_data.age_at(target_date)
    expected_age = 23.99726  # Approximately, not reaching the 24th birthday
    assert abs(age - expected_age) < 0.01, f"Expected age {expected_age}, got {age}"


def test_interpolate_fraction_out_of_bounds() -> None:
    # Even though the function does not restrict fraction, test behavior
    assert (
        interpolate(0, 10, -0.5) == -5.0
    ), "Interpolation failed for negative fraction"
    assert interpolate(0, 10, 1.5) == 15.0, "Interpolation failed for fraction >1"


# Tests for post-FIRE spending feature
def test_post_fire_spending_uses_lower_target_for_fire() -> None:
    """Test that FIRE is calculated based on post_fire_spending when set."""
    # Create input where current spending is high but target spending is low
    input_data = InputData(
        growth_rate=7.0,
        current_nw=500000.0,  # $500k
        spending_per_month=5000.0,  # $5k/month current spending
        inflation=2.0,
        annual_salary_increase=3.0,
        income_per_month=8000.0,
        extra_income=0.0,
        date_of_birth=datetime.date(1990, 1, 1),
        safe_withdraw_rate=4.0,
        post_fire_spending_per_month=2000.0,  # $2k/month target (cheaper country)
    )

    results = calculate_results_for_month(input_data)
    first_result = results[0]

    # Verify post_fire_spending is set
    assert first_result.post_fire_spending == 2000.0

    # fire_spending_target should return post_fire_spending when set
    assert first_result.fire_spending_target == 2000.0

    # safe_withdraw_minus_spending should compare against target, not current
    # With $500k and 4% SWR: yearly = $20k, monthly = $1666.67
    # This is less than $2000 target, so should be negative
    expected_swr_monthly = 500000.0 * 0.04 / 12
    expected_diff = expected_swr_monthly - 2000.0
    assert abs(first_result.safe_withdraw_minus_spending - expected_diff) < 0.01


def test_post_fire_spending_none_uses_current_spending() -> None:
    """Test that without post_fire_spending, current spending is used for FIRE."""
    input_data = InputData(
        growth_rate=7.0,
        current_nw=500000.0,
        spending_per_month=3000.0,
        inflation=2.0,
        annual_salary_increase=3.0,
        income_per_month=8000.0,
        extra_income=0.0,
        date_of_birth=datetime.date(1990, 1, 1),
        safe_withdraw_rate=4.0,
        post_fire_spending_per_month=None,  # No target - use current spending
    )

    results = calculate_results_for_month(input_data)
    first_result = results[0]

    # Verify post_fire_spending is None
    assert first_result.post_fire_spending is None

    # fire_spending_target should fall back to current spending
    assert first_result.fire_spending_target == 3000.0

    # safe_withdraw_minus_spending should compare against current spending
    expected_swr_monthly = 500000.0 * 0.04 / 12
    expected_diff = expected_swr_monthly - 3000.0
    assert abs(first_result.safe_withdraw_minus_spending - expected_diff) < 0.01


def test_post_fire_spending_earlier_fire_date() -> None:
    """Test that lower post_fire_spending leads to earlier FIRE date."""
    base_input = InputData(
        growth_rate=7.0,
        current_nw=200000.0,
        spending_per_month=4000.0,
        inflation=2.0,
        annual_salary_increase=3.0,
        income_per_month=7000.0,
        extra_income=0.0,
        date_of_birth=datetime.date(1990, 1, 1),
        safe_withdraw_rate=4.0,
    )

    # Calculate with current spending as target
    results_current = calculate_results_for_month(base_input)
    summary_current = Summary.from_results(results_current)

    # Calculate with lower target spending
    input_with_target = base_input.model_copy(
        update={"post_fire_spending_per_month": 2500.0}
    )
    results_target = calculate_results_for_month(input_with_target)
    summary_target = Summary.from_results(results_target)

    # Both should reach FIRE
    assert summary_current is not None
    assert summary_target is not None

    # Target spending version should reach FIRE earlier
    assert summary_target.fire_age < summary_current.fire_age
    assert summary_target.years_till_fi < summary_current.years_till_fi


def test_post_fire_spending_inflation_adjusted() -> None:
    """Test that post_fire_spending is adjusted for inflation over time."""
    input_data = InputData(
        growth_rate=7.0,
        current_nw=100000.0,
        spending_per_month=3000.0,
        inflation=2.0,
        annual_salary_increase=3.0,
        income_per_month=6000.0,
        extra_income=0.0,
        date_of_birth=datetime.date(1990, 1, 1),
        safe_withdraw_rate=4.0,
        post_fire_spending_per_month=2000.0,
    )

    results = calculate_results_for_month(input_data, target=12)

    # Check that post_fire_spending increases with inflation
    first = results[0]
    last = results[-1]

    assert first.post_fire_spending == 2000.0
    # After 12 months with 2% annual inflation, target should increase
    expected_target_after_12_months = 2000.0 * (input_data.monthly_inflation**12)
    assert abs(last.post_fire_spending - expected_target_after_12_months) < 0.01



from .fire import Results, Summary
from datetime import datetime

_PLOT_PROPERTIES = dict(width=360, usermeta={"embedOptions": {"actions": False}})


def _format_time_from_now(years: float) -> str:
    """Format years from now into readable text."""
    if years < 0.1:
        return "now"
    else:
        return f"{years:.1f} yr"


def _create_data_point(result, current_date, y_field: str):
    """Create a standardized data point with date, y-value, age, and time info."""
    time_years = (result.date - current_date).days / 365.25
    return {
        "x": result.date.isoformat(),
        "y": getattr(result, y_field),
        "age": result.age,
        "time_from_now": time_years,
        "time_from_now_text": _format_time_from_now(time_years),
    }


def _get_theme_colors(theme: str = "light") -> dict:
    """Get theme-appropriate colors for charts."""
    if theme == "dark":
        return {
            "background": "#2c2c2c",
            "foreground": "#ecf0f1",
            "gridColor": "#404040",
            "tooltipBg": "#404040",
            "tooltipColor": "#ecf0f1",
        }
    else:
        return {
            "background": "#ffffff",
            "foreground": "#2c3e50",
            "gridColor": "#e9ecef",
            "tooltipBg": "#ffffff",
            "tooltipColor": "#2c3e50",
        }


def _create_base_chart_config(
    chart_data: dict, y_axis_title: str, theme: str = "light"
) -> dict:
    """Create base ApexCharts configuration with theme support."""
    colors = _get_theme_colors(theme)

    config = {
        "series": chart_data["series"],
        "chart": {
            "type": "area",
            "stacked": False,
            "height": 350,
            "background": colors["background"],
            "foreColor": colors["foreground"],
            "zoom": {"type": "x", "enabled": True, "autoScaleYaxis": True},
            "toolbar": {"show": False},
            "animations": {"enabled": True, "easing": "easeinout", "speed": 800},
        },
        "dataLabels": {"enabled": False},
        "markers": {"size": 0},
        "fill": {
            "type": "gradient",
            "gradient": {
                "shadeIntensity": 1,
                "inverseColors": False,
                "opacityFrom": 0.5,
                "opacityTo": 0,
                "stops": [0, 90, 100],
            },
        },
        "colors": ["#ff6b35", "#6c5ce7", "#00b894", "#e17055", "#0984e3"],
        "stroke": {"curve": "smooth", "width": 2},
        "grid": {
            "borderColor": colors["gridColor"],
            "strokeDashArray": 0,
            "xaxis": {"lines": {"show": True}},
            "yaxis": {"lines": {"show": True}},
        },
        "xaxis": {
            "type": "datetime",
            "title": {"text": "Date", "style": {"color": colors["foreground"]}},
            "labels": {"style": {"colors": colors["foreground"]}},
        },
        "yaxis": {
            "title": {"text": y_axis_title, "style": {"color": colors["foreground"]}},
            "labels": {"style": {"colors": colors["foreground"]}},
        },
        "tooltip": {
            "theme": theme,
            "shared": True,
            "style": {"fontSize": "12px", "color": colors["tooltipColor"]},
        },
        "legend": {
            "position": "top",
            "horizontalAlign": "left",
            "labels": {"colors": colors["foreground"]},
        },
    }

    # Add FIRE date annotation if available
    if chart_data.get("fire_date"):
        # Convert ISO date string to timestamp (milliseconds since epoch) for ApexCharts
        fire_date_dt = datetime.fromisoformat(chart_data["fire_date"])
        fire_date_timestamp = int(fire_date_dt.timestamp() * 1000)

        config["annotations"] = {
            "xaxis": [
                {
                    "x": fire_date_timestamp,
                    "borderColor": colors["foreground"],
                    "strokeDashArray": 5,
                    "label": {
                        "text": "FIRE Date",
                        "style": {
                            "color": colors["foreground"],
                            "background": colors["background"],
                        },
                    },
                }
            ]
        }

    return config


def plot_age_vs_net_worth(results: list[Results], summary: Summary):
    """Generate complete ApexCharts configuration for net worth chart."""
    current_date = results[0].input_data.now

    # Create data series using the helper function
    series_data = [
        ("Net Worth", "nw"),
        ("Saved", "total_saved"),
        ("Profits", "total_investment_profits"),
    ]

    series = []
    for name, field in series_data:
        data = [_create_data_point(result, current_date, field) for result in results]
        series.append({"name": name, "data": data})

    chart_data = {
        "series": series,
        "fire_date": summary.fire_date.isoformat(),
    }

    return {
        "config_light": _create_base_chart_config(chart_data, "Amount ($)", "light"),
        "config_dark": _create_base_chart_config(chart_data, "Amount ($)", "dark"),
    }


def plot_age_vs_monthly_safe_withdraw(results: list[Results], summary: Summary):
    """Generate data for ApexCharts monthly safe withdrawal chart."""
    monthly_safe_withdraw_data = [
        {"x": result.date.isoformat(), "y": result.safe_withdraw_rule_monthly}
        for result in results
    ]
    spending_data = [
        {"x": result.date.isoformat(), "y": result.spending} for result in results
    ]
    income_data = [
        {"x": result.date.isoformat(), "y": result.income} for result in results
    ]

    return {
        "series": [
            {"name": "Monthly Safe Withdraw", "data": monthly_safe_withdraw_data},
            {"name": "Spending", "data": spending_data},
            {"name": "Income", "data": income_data},
        ],
        "fire_age": summary.fire_age,
        "fire_date": summary.fire_date.isoformat(),
        "spending_at_fi": summary.spending_at_fi,
    }


def plot_savings_vs_spending(results: list[Results], summary: Summary):
    """Generate data for ApexCharts savings vs spending chart."""
    savings_data = [
        {"x": result.date.isoformat(), "y": result.saving} for result in results
    ]
    spending_data = [
        {"x": result.date.isoformat(), "y": result.spending} for result in results
    ]
    investment_profits_data = [
        {"x": result.date.isoformat(), "y": result.investment_profits}
        for result in results
    ]

    return {
        "series": [
            {"name": "Savings", "data": savings_data},
            {"name": "Spending", "data": spending_data},
            {"name": "Investment Profits", "data": investment_profits_data},
        ],
        "fire_age": summary.fire_age,
        "fire_date": summary.fire_date.isoformat(),
        "saving_at_fi": summary.saving_at_fi,
        "spending_at_fi": summary.spending_at_fi,
    }


def plot_monthly_financial_flows(results: list[Results], summary: Summary):
    """Generate complete ApexCharts configuration for monthly financial flows chart."""
    current_date = results[0].input_data.now

    # Create data series using the helper function
    # Use actual_spending which switches to post-FIRE spending after FIRE is reached
    series_data = [
        ("Monthly Safe Withdraw", "safe_withdraw_rule_monthly"),
        ("Income", "income"),
        ("Spending", "actual_spending"),
        ("Savings", "saving"),
        ("Investment Profits", "investment_profits"),
    ]

    series = []
    for name, field in series_data:
        data = [_create_data_point(result, current_date, field) for result in results]
        series.append({"name": name, "data": data})

    chart_data = {
        "series": series,
        "fire_date": summary.fire_date.isoformat(),
    }

    return {
        "config_light": _create_base_chart_config(
            chart_data, "Monthly Amount ($)", "light"
        ),
        "config_dark": _create_base_chart_config(
            chart_data, "Monthly Amount ($)", "dark"
        ),
    }



"""WenFire - Financial Independence Calculator."""

try:
    from ._version import __version__, __version_tuple__
except ImportError:
    __version__ = "0.0.0+unknown"
    __version_tuple__ = (0, 0, 0, "+unknown")

__all__ = ["__version__", "__version_tuple__"]
